""" Utility functions. """

from __future__ import absolute_import
from celery.utils.log import get_task_logger
from sailthru.sailthru_client import SailthruClient

from ecommerce_worker.sailthru.v1.exceptions import ConfigurationError, SailthruNotEnabled
from ecommerce_worker.utils import get_configuration

log = get_task_logger(__name__)


def get_sailthru_configuration(site_code):
    """ Returns the Sailthru configuration for the specified site. """
    config = get_configuration('SAILTHRU', site_code=site_code)
    return config


def get_sailthru_client(site_code):
    """
    Returns a Sailthru client for the specified site.

    Args:
        site_code (str): Site for which the client should be configured.

    Returns:
        SailthruClient

    Raises:
        SailthruNotEnabled: If Sailthru is not enabled for the specified site.
        ConfigurationError: If either the Sailthru API key or secret are not set for the site.
    """
    # Get configuration
    config = get_sailthru_configuration(site_code)

    # Return if Sailthru integration disabled
    if not config.get('SAILTHRU_ENABLE'):
        msg = 'Sailthru is not enabled for site {}'.format(site_code)
        log.debug(msg)
        raise SailthruNotEnabled(msg)

    # Make sure key and secret configured
    key = config.get('SAILTHRU_KEY')
    secret = config.get('SAILTHRU_SECRET')

    if not (key and secret):
        msg = 'Both key and secret are required for site {}'.format(site_code)
        log.error(msg)
        raise ConfigurationError(msg)

    return SailthruClient(key, secret)


def can_retry_sailthru_request(error):
    """ Returns True if a Sailthru request and be re-submitted after an error has occurred.

    Responses with the following codes can be retried:
         9: Internal Error
        43: Too many [type] requests this minute to /[endpoint] API

    All other errors are considered failures, that should not be retried. A complete list of error codes is available at
    https://getstarted.sailthru.com/new-for-developers-overview/api/api-response-errors/.

    Args:
        error (SailthruResponseError)

    Returns:
        bool: Indicates if the original request can be retried.
    """
    code = error.get_error_code()
    return code in (9, 43)



""" Tests for utility functions. """

from __future__ import absolute_import
from unittest import TestCase

import ddt
import mock

from ecommerce_worker.sailthru.v1.exceptions import SailthruNotEnabled, ConfigurationError
from ecommerce_worker.sailthru.v1.utils import get_sailthru_client


@ddt.ddt
class SailthruUtilsTests(TestCase):
    """ Tests for Sailthru utility functions. """
    SITE_CODE = 'test'
    SITE_OVERRIDES_MODULE = 'ecommerce_worker.configuration.test.SITE_OVERRIDES'

    def assert_get_sailthru_client_raises(self, exc_class, config):
        """ Asserts an error is raised by a call to get_sailthru_client. """
        overrides = {
            self.SITE_CODE: {
                'SAILTHRU': config
            }
        }
        with mock.patch.dict(self.SITE_OVERRIDES_MODULE, overrides):
            with self.assertRaises(exc_class):
                get_sailthru_client(self.SITE_CODE)

    def test_get_sailthru_client_with_sailthru_disabled(self):
        """ Verify the method raises a SailthruNotEnabled if Sailthru is not enabled for the site. """

        with mock.patch('ecommerce_worker.sailthru.v1.utils.log.debug') as mock_log:
            self.assert_get_sailthru_client_raises(SailthruNotEnabled, {'SAILTHRU_ENABLE': False})

        mock_log.assert_called_once_with('Sailthru is not enabled for site {}'.format(self.SITE_CODE))

    @ddt.data(
        {},
        {'SAILTHRU_KEY': None, 'SAILTHRU_SECRET': None},
        {'SAILTHRU_KEY': 'test', 'SAILTHRU_SECRET': None},
        {'SAILTHRU_KEY': None, 'SAILTHRU_SECRET': 'test'},
    )
    def test_get_sailthru_client_without_credentials(self, sailthru_config):
        """ Verify the method raises a ConfigurationError if Sailthru is not configured properly. """
        sailthru_config['SAILTHRU_ENABLE'] = True

        with mock.patch('ecommerce_worker.sailthru.v1.utils.log.error') as mock_log:
            self.assert_get_sailthru_client_raises(ConfigurationError, sailthru_config)

        mock_log.assert_called_once_with('Both key and secret are required for site {}'.format(self.SITE_CODE))

    def test_get_sailthru_client(self):
        """ Verify the method returns an authenticated SailthruClient. """
        key = 'key'
        secret = 'secret'
        overrides = {
            self.SITE_CODE: {
                'SAILTHRU': {
                    'SAILTHRU_ENABLE': True,
                    'SAILTHRU_KEY': key,
                    'SAILTHRU_SECRET': secret
                }
            }
        }
        with mock.patch.dict(self.SITE_OVERRIDES_MODULE, overrides):
            client = get_sailthru_client(self.SITE_CODE)
            self.assertEqual(client.api_key, key)
            self.assertEqual(client.secret, secret)



"""
    Braze Client functions.
"""

from __future__ import absolute_import

import requests

from urllib.parse import urlencode

from celery.utils.log import get_task_logger

from ecommerce_worker.braze.v1.exceptions import (
    ConfigurationError,
    BrazeNotEnabled,
    BrazeClientError,
    BrazeRateLimitError,
    BrazeInternalServerError
)
from ecommerce_worker.utils import get_configuration

log = get_task_logger(__name__)


def get_braze_configuration(site_code):
    """
    Returns the Braze configuration for the specified site.

    Arguments:
        site_code (str): Site for fetching the configuration.
    """
    config = get_configuration('BRAZE', site_code=site_code)
    return config


def get_braze_client(site_code, endpoint=None):
    """
    Returns a Braze client for the specified site.

    Arguments:
        site_code (str): Site for which the client should be configured.
        endpoint (str): The endpoint for the API e.g. /messages/send or /email/hard_bounces

    Returns:
        BrazeClient

    Raises:
        BrazeNotEnabled: If Braze is not enabled for the specified site.
        ConfigurationError: If either the Braze API key or Webapp key are not set for the site.
    """
    # Get configuration
    config = get_braze_configuration(site_code)

    # Return if Braze integration disabled
    if not config.get('BRAZE_ENABLE'):
        msg = 'Braze is not enabled for site {}'.format(site_code)
        log.debug(msg)
        raise BrazeNotEnabled(msg)

    rest_api_key = config.get('BRAZE_REST_API_KEY')
    webapp_api_key = config.get('BRAZE_WEBAPP_API_KEY')
    rest_api_url = config.get('REST_API_URL')
    message_endpoint = endpoint or config.get('MESSAGES_SEND_ENDPOINT')
    from_email = config.get('FROM_EMAIL')

    if (
            not rest_api_key or
            not webapp_api_key or
            not rest_api_url or
            not message_endpoint or
            not from_email
    ):
        msg = 'Required parameters missing for site {}'.format(site_code)
        log.error(msg)
        raise ConfigurationError(msg)

    return BrazeClient(
        rest_api_key=rest_api_key,
        webapp_api_key=webapp_api_key,
        request_url=rest_api_url + message_endpoint,
        from_email=from_email
    )


class BrazeClient(object):
    """
    Client for Braze REST API
    """

    def __init__(self, rest_api_key, webapp_api_key, request_url, from_email):
        """
        Initialize the Braze Client with configuration values.

        Arguments:
            rest_api_key (str): API key to use for accessing Braze REST endpoints
            webapp_api_key (str): Key that identifies the Braze webapp application
            request_url (str): Braze endpoint url
            from_email (str): Braze email from address
        """
        self.rest_api_key = rest_api_key
        self.webapp_api_key = webapp_api_key
        self.from_email = from_email
        self.session = requests.Session()
        self.request_url = request_url

    def __create_post_request(self, body):
        """
        Creates a request and returns a response.

        Arguments:
            body (dict): The request body

        Returns:
            response (dict): The response object
        """
        response = {'errors': []}
        r = self._post_request(body)
        response.update(r.json())
        response['status_code'] = r.status_code

        message = response["message"]
        response['success'] = (
            message in ('success', 'queued') and not response['errors']
        )
        if message not in ('success', 'queued'):
            raise BrazeClientError(message, response['errors'])
        return response

    def _post_request(self, body):
        """
        Http posts the message body with associated headers.

        Arguments:
            body (dict): The request body

        Returns:
            r (requests.Response): The http response object
        """
        self.session.headers.update(
            {'Authorization': u'Bearer {}'.format(self.rest_api_key), 'Content-Type': 'application/json'}
        )
        r = self.session.post(self.request_url, json=body, timeout=2)
        if r.status_code == 429:
            reset_epoch_s = float(r.headers.get('X-RateLimit-Reset', 0))
            raise BrazeRateLimitError(reset_epoch_s)
        elif str(r.status_code).startswith('5'):
            raise BrazeInternalServerError
        return r

    def __create_get_request(self, parameters):
        """
        Creates a request and returns a response.

        Arguments:
            parameters (dict): The request parameters

        Returns:
            response (dict): The response object
        """
        response = {'errors': []}
        r = self._get_request(parameters)
        response.update(r.json())
        response['status_code'] = r.status_code
        message = response["message"]
        response['success'] = (
            message in ('success', 'queued') and not response['errors']
        )
        if message not in ('success', 'queued'):
            raise BrazeClientError(message, response['errors'])
        return response

    def _get_request(self, parameters):
        """
        Http GET the parameters with associated headers.

        Arguments:
            parameters (dict): The request parameters

        Returns:
            r (requests.Response): The http response object
        """
        self.session.headers.update(
            {'Authorization': u'Bearer {}'.format(self.rest_api_key), 'Content-Type': 'application/json'}
        )
        url_with_parameters = self.request_url + '?' + urlencode(parameters)
        r = self.session.get(url_with_parameters)
        if r.status_code == 429:
            reset_epoch_s = float(r.headers.get('X-RateLimit-Reset', 0))
            raise BrazeRateLimitError(reset_epoch_s)
        elif str(r.status_code).startswith('5'):
            raise BrazeInternalServerError
        return r

    def send_message(
        self,
        email_ids,
        subject,
        body,
        sender_alias='EdX Support Team'
    ):
        """
        Sends the message via Braze Rest API /messages/send

        Arguments:
            email_ids (list): e.g. ['test1@example.com', 'test2@example.com']
            subject (str): e.g. 'Test Subject'
            body (str): e.g. '<html>Test Html Message</html>'
            sender_alias (str): sender alias for email e.g. edX Support Team

        Request message format:
            Content-Type: application/json
            Authorization: Bearer {{api_key}}
            Body:
                {
                    "external_user_ids": [ "user1@example.com", "user2@example.org" ],
                    "messages": {
                        "email": {
                            "app_id": "99999999-9999-9999-9999-999999999999",
                            "subject": "Testing",
                            "from": "edX <edx-for-business-no-reply@info.edx.org>",
                            "body": "<html>Test</html>"
                        }
                    }
                }

        Returns:
            response (dict): e.g:
                {
                    "dispatch_id": "9179894df02dab49c1d1f85c6a01c41b",
                    "message": "success"
                }
        """
        if not email_ids or not subject or not body:
            raise BrazeClientError("Missing parameters for Braze email")
        email = {
            'app_id': self.webapp_api_key,
            'subject': subject,
            'from': sender_alias + self.from_email,
            'body': body
        }
        message = {
            'external_user_ids': email_ids,
            'messages': {
                'email': email
            }
        }

        return self.__create_post_request(message)

    def did_email_bounce(
        self,
        email_id
    ):
        """
        Sends the message via Braze Rest API /email/hard_bounces

        Arguments:
            email_id (str): e.g. 'test1@example.com'

        Returns:
             True (boolean): True if email bounced,
        """
        if not email_id:
            raise BrazeClientError("Missing parameters for Braze email")
        parameters = {
            'email': email_id
        }

        response = self.__create_get_request(parameters)
        if response['emails']:
            return True

        return False



""" Test coverage for ecommerce_worker/utils.py """
from __future__ import absolute_import
from unittest import TestCase

import ddt
import mock

from ecommerce_worker.configuration.test import ECOMMERCE_API_ROOT
from ecommerce_worker.utils import get_configuration


@ddt.ddt
class GetConfigurationTests(TestCase):
    """Tests covering the get_configuration operation."""

    SITE_OVERRIDES_MODULE = 'ecommerce_worker.configuration.test.SITE_OVERRIDES'
    TEST_SETTING = 'ECOMMERCE_API_ROOT'

    OVERRIDES_DICT = {
        'openedx': {'ECOMMERCE_API_ROOT': 'http://openedx.org'},
        'test': {'ECOMMERCE_API_ROOT': 'http://example.com'},
        'novalue': {'ECOMMERCE_API_ROOT': None},
        'emptydict': {}
    }

    def test_configuration_no_site_overrides(self):
        test_setting = get_configuration(self.TEST_SETTING)
        self.assertEqual(test_setting, ECOMMERCE_API_ROOT)

    def test_configuration_no_site_code_specified(self):
        with mock.patch.dict(self.SITE_OVERRIDES_MODULE, self.OVERRIDES_DICT):
            test_setting = get_configuration(self.TEST_SETTING)
            self.assertEqual(test_setting, ECOMMERCE_API_ROOT)

    @ddt.data('openedx', 'test')
    def test_configuration_valid_site_code_dict_value(self, site_code):
        """
        Confirm that valid SITE_OVERRIDES parameters are correctly returned
        """
        with mock.patch.dict(self.SITE_OVERRIDES_MODULE, self.OVERRIDES_DICT):
            test_setting = get_configuration(self.TEST_SETTING, site_code=site_code)
            self.assertEqual(test_setting, self.OVERRIDES_DICT[site_code][self.TEST_SETTING])

    @ddt.data(None, 'invalid', 'emptydict', 'novalue')
    def test_configuration_no_site_code_matched(self, site_code):
        """
        Test various states of SITE_OVERRIDES to ensure all branches (ie, use cases) are covered
        """
        with mock.patch.dict(self.SITE_OVERRIDES_MODULE, self.OVERRIDES_DICT):
            test_setting = get_configuration(self.TEST_SETTING, site_code=site_code)
            self.assertEqual(test_setting, ECOMMERCE_API_ROOT)
